
/*MA****************************************************************/
/*                                                                 */
/*     File: tree.C                                                */
/*   Author: Helmut Schmid                                         */
/*  Purpose:                                                       */
/*  Created: Tue Jan 21 13:27:26 2003                              */
/* Modified: Fri Nov 13 08:33:34 2009 (schmid)                     */
/*                                                                 */
/*ME****************************************************************/

#include <stdio.h>

#include <iostream>
using std::cerr;


#include "tree.h"

bool Verbose=true;
bool PrintRules=false;
bool WithHeads=false;
bool PrettyPrint=false;
bool PrintTraces=false;
bool UndoMarkov=false;
bool EmptyCategories=false;
bool ConvertTraces=false;

char *MapFileName=NULL;
char *PrepFileName=NULL;
SymbolMap CatMap;
SymbolMap PrepMap;

FILE *govfile=NULL, *outfile=NULL, *lexfile=NULL;

#include "symbol-table.h"

SymbolTable ST;

typedef enum {left, right} Dir;

typedef struct {
  size_t catnum;
  Dir dir;
  vector<size_t> heads;
} HeadRule;

vector<HeadRule> HeadRules;


/*FA****************************************************************/
/*                                                                 */
/*  is_year_number                                                 */
/*                                                                 */
/*FE****************************************************************/

bool is_year_number( char *l )

{
  return (((l[0] == '1' && l[1] == '9' && l[2] >= '0' && l[2] <= '9') ||
	   (l[0] == '2' && l[1] == '0' && l[2] >= '0' && l[2] <= '1')) &&
	  l[3] >= '0' && l[3] <= '9' && l[4] == 0);
}


/*FA****************************************************************/
/*                                                                 */
/*  is_month_name                                                  */
/*                                                                 */
/*FE****************************************************************/

bool is_month_name( char *l )

{
  return (strcmp(l,"January") == 0 || strcmp(l,"February") == 0 ||
	  strcmp(l,"March") == 0 || strcmp(l,"April") == 0 ||
	  strcmp(l,"May") == 0 || strcmp(l,"June") == 0 ||
	  strcmp(l,"July") == 0 || strcmp(l,"August") == 0 ||
	  strcmp(l,"September") == 0 || strcmp(l,"October") == 0 ||
	  strcmp(l,"November") == 0 || strcmp(l,"December") == 0);
}


/*FA****************************************************************/
/*                                                                 */
/*  is_day_of_week                                                 */
/*                                                                 */
/*FE****************************************************************/

bool is_day_of_week( char *l )

{
  return (strcmp(l,"Monday") == 0 || strcmp(l,"Tuesday") == 0 ||
	  strcmp(l,"Wednesday") == 0 || strcmp(l,"Thursday") == 0 ||
	  strcmp(l,"Friday") == 0 || strcmp(l,"Saturday") == 0 ||
	  strcmp(l,"Sunday") == 0);
}
    

/*FA****************************************************************/
/*                                                                 */
/*  read_head_rules                                                */
/*                                                                 */
/*FE****************************************************************/

void read_head_rules( char *filename )

{
  FILE *file;
  if ((file = fopen(filename,"rt")) == NULL) {
    cerr << "Error: unable to open map file \"" << filename << "\"!\n";
    exit(1);
  }
  char buffer[10000];
  while (fgets(buffer,10000,file)) {
    HeadRule r;
    char *p=buffer;
    char *cat=strtok_r(p," \t",&p);
    r.catnum = ST.number(cat);

    char *dir=strtok_r(p," \t\n",&p);
    if (strcmp(dir,"left") == 0)
      r.dir = left;
    else if (strcmp(dir,"right") == 0)
      r.dir = right;
    else {
      cerr << "Error: not a direction: \"" << dir << "\"\n";
      exit(1);
    }

    char *head;
    while ((head = strtok_r(p," \t\n",&p))) {
      size_t catnum = ST.number(head);
      r.heads.push_back(catnum);
    }

    HeadRules.push_back(r);
  }
  fclose(file);
}


/*FA****************************************************************/
/*                                                                 */
/*  lookup_cat                                                     */
/*                                                                 */
/*FE****************************************************************/

static size_t lookup_cat( char *label )

{
  char buffer[1000];
  int i;
  for( i=0; label[i]; i++ )
    if (i > 0 && label[i] == '-' && label[i+1] != 0)
      break;
    else
      buffer[i] = label[i];
  buffer[i] = 0;

  return ST.number(buffer);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::assign_heads                                             */
/*                                                                 */
/*FE****************************************************************/

void Tree::assign_heads()

{
  catnum = lookup_cat( label );

  if (daughters.size() == 1 && daughters[0]->daughters.size() == 0) {
    head_daughter = NULL;
    if (matching_category("-NONE-"))
      head = NULL;
    else
      head = this;
    return;
  }

  for( size_t d=0; d<daughters.size(); d++ )
    daughters[d]->assign_heads();

  for( size_t i=0; i<HeadRules.size(); i++ ) {
    HeadRule &r=HeadRules[i];
    if (r.catnum == catnum) {
      vector<size_t> &cats=r.heads;
      if (r.dir == left) {
	for( size_t d=0; d<daughters.size(); d++ )
	  if (daughters[d]->head != NULL) {
	    size_t l;
	    for( l=0; l<cats.size(); l++ )
	      if (cats[l] == daughters[d]->catnum)
		break;
	    if (cats.size() == 0 || l < cats.size()) {
	      head_daughter = daughters[d];
	      head = daughters[d]->head;
	      return;
	    }
	  }
      }
      else {  // r.dir == right
	int last=(int)daughters.size()-1;
	for( int d=last; d>=0; d-- ) {
	  if (daughters[d]->head != NULL) {
	    size_t l;
	    for( l=0; l<cats.size(); l++ )
	      if (cats[l] == daughters[d]->catnum)
		break;
	    if (cats.size() == 0 || l < cats.size()) {
	      head_daughter = daughters[d];
	      head = daughters[d]->head;
	      return;
	    }
	  }
	}
      }
    }
  }
  // general default
  for( size_t d=0; d<daughters.size(); d++ )
    if (daughters[d]->head != NULL) {
      head_daughter = daughters[d];
      head = daughters[d]->head;
      return;
    }
  head_daughter = daughters[0];
  head = daughters[0]->head;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::strip_features                                           */
/*                                                                 */
/*FE****************************************************************/

void Tree::strip_features()

{
  if (daughters.size() == 0)
    return;
  for( char *p=label; *p; p++ )
    if (*p == '/') {
      *p = 0;
      break;
    }
}


/*FA****************************************************************/
/*                                                                 */
/*  last_char                                                      */
/*                                                                 */
/*FE****************************************************************/

char *last_char( char *s ) {
  if (*s) {
    do { s++; } while (*s);
    s--;
  }
  return s;
}

const char *last_char( const char *s ) {
  if (*s) {
    do { s++; } while (*s);
    s--;
  }
  return s;
}

bool is_digit( char c ) { return c >= '0' && c <= '9'; }


/*FA****************************************************************/
/*                                                                 */
/*  read_mapping                                                   */
/*                                                                 */
/*  Purpose: reads a mapping from Penn tags to (simpler) Penn tags */
/*           from a file                                           */
/*                                                                 */
/*FE****************************************************************/

void read_mapping( char *filename, SymbolMap &map )

{
  FILE *file;
  if ((file = fopen(filename,"rt")) == NULL) {
    cerr << "Error: unable to open map file \"" << filename << "\"!\n";
    exit(1);
  }
  char buffer[1000];
  while (fgets(buffer,1000,file)) {
    char *s1=strtok(buffer," \t\r");
    char *s2=strtok(NULL," \t\r\n");
    if (s2)
      map[s1] = strdup(s2);
  }
  fclose(file);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::predecessor                                              */
/*                                                                 */
/*FE****************************************************************/

Tree *Tree::predecessor()

{
  if (mother)
    for( size_t i=1; i<mother->daughters.size(); i++ )
      if (mother->daughters[i] == this)
	return mother->daughters[i-1];
  
  return NULL;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::successor                                                */
/*                                                                 */
/*FE****************************************************************/

Tree *Tree::successor()

{
  if (mother)
    for( size_t i=0; i<mother->daughters.size()-1; i++ )
      if (mother->daughters[i] == this)
	return mother->daughters[i+1];
  
  return NULL;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::is_first_daughter                                        */
/*                                                                 */
/*FE****************************************************************/

bool Tree::is_first_daughter()

{
  return (mother && this == mother->daughters[0]);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::is_last_daughter                                         */
/*                                                                 */
/*FE****************************************************************/

bool Tree::is_last_daughter()

{
  return (mother && this == mother->last_daughter());
}


/*FA****************************************************************/
/*                                                                 */
/*  index_start                                                    */
/*                                                                 */
/*FE****************************************************************/

bool index_start( const char *s )

{
  return *(s++) == '-' && is_digit(*s);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::noun_POS_tag                                             */
/*                                                                 */
/*FE****************************************************************/

bool Tree::noun_POS_tag()

{
  return matching_category("NN") || matching_category("NNS") ||
    matching_category("NNP") || matching_category("NNPS");
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::verb_POS_tag                                             */
/*                                                                 */
/*FE****************************************************************/

bool Tree::verb_POS_tag()

{
  return (matching_category("V?") && !matching_category("VP"))
    || matching_category("V??");
}


/*FA****************************************************************/
/*                                                                 */
/*  feature_boundary                                               */
/*                                                                 */
/*FE****************************************************************/

static bool feature_boundary( const char *p )

{
  return (*p == '_' || *p == '+' || *p == '=' || *p == '/' || 
	  (*p == '%' && *(p-1) != '/') || index_start(p));
}


/*FA****************************************************************/
/*                                                                 */
/*  slash_feature_boundary                                         */
/*                                                                 */
/*FE****************************************************************/

static bool slash_feature_boundary( const char *p )

{
  return (*p == '_' || *p == '+' || *p == '=' ||
	  (*p == '%' && *(p-1) != '/') || index_start(p));
}


/*FA****************************************************************/
/*                                                                 */
/*  split_catname                                                  */
/*                                                                 */
/*  Purpose: splits category label into Penn tag and features      */
/*                                                                 */
/*FE****************************************************************/

static void split_catname( char *s, char **r1, char **r2 )

{
  static char buffer1[1000];
  static char buffer2[1000];

  int i;
  for( i=0; s[i]; i++ )
    if ((s[i] >= 'A' && s[i] <= 'Z') || 
	(s[i] == '-' && s[i+1] >= 'A' && s[i+1] <= 'Z'))
      buffer1[i] = s[i];
    else
      break;
  buffer1[i] = 0;
  s += i;
  for( i=0; s[i]; i++ )
    buffer2[i] = s[i];
  buffer2[i] = 0;
  *r1 = buffer1;
  *r2 = buffer2;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::category                                                 */
/*                                                                 */
/*FE****************************************************************/

char *Tree::category()

{
  static char buffer[1000];

  int i;
  for( i=0; label[i] && !feature_boundary(label+i); i++ )
    buffer[i] = label[i];
  buffer[i] = 0;

  return buffer;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::category2                                                */
/*                                                                 */
/*FE****************************************************************/

char *Tree::category2()

{
  char *s = category();

  // cut off Penn-Tags like -TMP
  int i;
  for( i=0; s[i]; i++ )
    if (s[i] == '-' && s[i+1]) {
      s[i] = 0;
      break;
    }

  return s;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::matching_category                                        */
/*                                                                 */
/*FE****************************************************************/

bool Tree::matching_category( const char *c )

{
  for( char *p=label; *p; p++ )
    if (*c == 0)
      return feature_boundary(p);
    else if (*p == *c || *c == '?')
      c++;
    else
      return false;
  return *c == 0;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::matching_base_cat                                        */
/*                                                                 */
/*FE****************************************************************/

bool Tree::matching_base_cat( const char *c )

{
  for( char *p=label; *p; p++, c++ ) {
    if (*c == 0)
      return (*p == '-' || feature_boundary(p));
    if (*p != *c && *c != '?')
      return false;
  }
  return *c == 0;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::matching_suffix                                          */
/*                                                                 */
/*FE****************************************************************/

bool Tree::matching_suffix( const char *suffix )

{
  char *l;
  for( l=label; *l; l++ ) ;
  const char *s;
  for( s=suffix; *s; s++ ) ;

  for( ; s>=suffix; s--, l--)
    if (l < label || *s != *l)
      return false;
  return true;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::is_preposition                                           */
/*                                                                 */
/*FE****************************************************************/

bool Tree::is_preposition()

{
  return matching_category("IN") || matching_category("TO");
}


/*FA****************************************************************/
/*                                                                 */
/*  append_string                                                  */
/*                                                                 */
/*FE****************************************************************/

static void append_string( char *&p, const char *q )

{
  while ((*(p++) = *(q++))) ;
  p--;
}


/*FA****************************************************************/
/*                                                                 */
/*  lowercase                                                      */
/*                                                                 */
/*FE****************************************************************/

static char *lowercase( const char *word )

{
  static char buffer[1000];
  for( int i=0; i<1000; i++) {
    if ((word[i] >= 'A' && word[i] <= 'Z') || 
	(word[i] >= '' && word[i] <= ''))
      buffer[i] = word[i] + 32;
    else
      buffer[i] = word[i];
    if (word[i] == 0)
      return buffer;
  }
  throw "Error: word too long!\n";
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::change_label                                             */
/*                                                                 */
/*FE****************************************************************/

void Tree::change_label( const char *word )

{
  char *s = label;
  label = strdup(word);
  free(s);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::base_cat                                                 */
/*                                                                 */
/*FE****************************************************************/

const char *Tree::base_cat()

{
  if (matching_base_cat("VP"))
    return "V";
  if (matching_base_cat("ADJP") ||
      matching_base_cat("WHADJP"))
    return "ADJ";
  if (matching_base_cat("ADVP") ||
      matching_base_cat("WHADVP"))
    return "ADV";
  if (matching_base_cat("NP") ||
      matching_base_cat("WHNP") ||
      matching_base_cat("QP"))
    return "NP";
  if (matching_base_cat("PP"))
    return "PP";
  return "0";
}


/*FA****************************************************************/
/*                                                                 */
/*  verb_type                                                      */
/*                                                                 */
/*FE****************************************************************/

static char verb_type( const char *word )

{
  char *w=lowercase(word);

  if (strcmp(w, "be") == 0 ||
      strcmp(w, "been") == 0 ||
      strcmp(w, "being") == 0 ||
      strcmp(w, "am") == 0 ||
      strcmp(w, "'m") == 0 ||
      strcmp(w, "is") == 0 ||
      strcmp(w, "'s") == 0 ||
      strcmp(w, "are") == 0 ||
      strcmp(w, "'re") == 0 ||
      strcmp(w, "were") == 0 ||
      strcmp(w, "was") == 0)
    return 'B';
  if (strcmp(w, "do") == 0 ||
      strcmp(w, "done") == 0 ||
      strcmp(w, "doing") == 0 ||
      strcmp(w, "does") == 0 ||
      strcmp(w, "did") == 0)
    return 'D';
  if (strcmp(w, "have") == 0 ||
      strcmp(w, "'ve") == 0 ||
      strcmp(w, "had") == 0 ||
      strcmp(w, "'d") == 0 ||
      strcmp(w, "having") == 0 ||
      strcmp(w, "has") == 0)
    return 'H';
  return 'V';
}


/*FA****************************************************************/
/*                                                                 */
/*  find_index                                                     */
/*                                                                 */
/*FE****************************************************************/

static char *find_index( char *label )

{
  for( char *p=label; *p; p++ )
    if (index_start(p)) {
      char *q;
      for( q=p+2; *q; q++ )
	if (!is_digit(*q))
	  break;
      if (*q == 0)
	return p;
    }

  return NULL;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::is_trace                                                 */
/*                                                                 */
/*FE****************************************************************/

bool Tree::is_trace()

{
  if (daughters.size() == 0 && mother)
    return mother->matching_category("-NONE-");

  for( size_t i=0; i<daughters.size(); i++ )
    if (!daughters[i]->is_trace())
      return false;

  return true;
}



/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_markup                                               */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_markup( const char *f, char d )

{
  char buffer[1000];
  char *l=label, *r=buffer;
  for(;;) {
    if (*f && (*l == 0 || slash_feature_boundary(l)))
      {
	*(r++) = d;
	while (*f)
	  *(r++) = *(f++) ;
	while ((*(r++) = *(l++))) ;
	break;
      }
    if ((*(r++) = *(l++)) == 0)
      break;
  }
  free(label);
  label = strdup(buffer);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::insert_feature                                           */
/*                                                                 */
/*FE****************************************************************/

void Tree::insert_feature( const char *f )

{
  add_markup(f, '/');
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::insert_uniq_feature                                      */
/*                                                                 */
/*FE****************************************************************/

void Tree::insert_uniq_feature( const char *f )

{
  for( char *p=label; *p; p++ )
    if (*p == '/')
      for( int i=0; ; i++ )
	if (f[i] == 0 && (p[i+1] == 0 || feature_boundary(p+1)))
	  return; // feature was previously added
	else if (f[i] != p[i+1])
	  break;
  
  add_markup(f, '/');
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::cut_off_indices                                          */
/*                                                                 */
/*FE****************************************************************/

const char *Tree::cut_off_indices()

{
  for( char *p=label; *p; p++ )
    if ((*p == '-'  || *p == '=') && is_digit(*(p+1))) {
      char *e;
      for( e=p+2; *e; e++ )
	if (is_digit(*e))
	  ;
	else if ((*e == '-' || *e == '=') && is_digit(*(e+1)))
	  e++;
	else
	  break;
      if (*e == 0) {
	static char buffer[1000];
	for( int i=0; (buffer[i] = p[i]); i++ );
	*p = 0;
	return buffer;
      }
    }
  return "";
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::append_slash                                             */
/*                                                                 */
/*FE****************************************************************/

void Tree::append_slash( char *prefix, char c, bool terminal )

{
  char buffer[1000];
  const char *indices=cut_off_indices();
  if (terminal)
    sprintf(buffer,"%s%c#%s%s",label,c,prefix,indices);
  else
    sprintf(buffer,"%s%c%s%s",label,c,prefix,indices);
  change_label(buffer);
}



/*FA****************************************************************/
/*                                                                 */
/*  quote                                                          */
/*                                                                 */
/*FE****************************************************************/

static char *quote( const char *string )

{
  int i,k;
  static char buffer[10000];
  bool is_trace=false;

  if (string[0] == '(' && *last_char(string) == ')')
    is_trace = true;
  
  for( i=k=0; k<10000; i++, k++) {
    if (!string[i])
      break;
    if (string[i] == '"' || string[i] == '\\'|| string[i] == ' ' || 
	string[i] == '$' || string[i] == '^' || string[i] == '\''||
	string[i] == '[' || string[i] == ']' || string[i] == '{' ||
	string[i] == '}' || string[i] == '=' || string[i] == '<' || 
	string[i] == '>' || string[i] == '#' || string[i] == '/' || 
	(!is_trace && (string[i] == '(' || string[i] == ')')))
      buffer[k++] = '\\';
    buffer[k] = string[i];
  }
  buffer[k] = '\0';

  return &buffer[0];
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::print_subtree                                            */
/*                                                                 */
/*FE****************************************************************/

void Tree::print_subtree( int depth )

{
  if (daughters.size() == 0)
    fprintf(outfile," %s", quote(label));
  else {
    if (PrettyPrint) {
      fputc('\n', outfile);
      for( int i=0; i<depth; i++ )
	fputs("  ", outfile);
    }
    fprintf(outfile, "(%s", quote(label));
    for( size_t i=0; i<daughters.size(); i++ )
      daughters[i]->print_subtree(depth+1);
    fputc(')', outfile);
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::print                                                    */
/*                                                                 */
/*FE****************************************************************/

void Tree::print()

{
  print_subtree(0);
  fputc('\n', outfile);
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::matching_trace                                           */
/*                                                                 */
/*FE****************************************************************/

char Tree::matching_trace( const char *suffix )

{
  // find trace index
  for( char *p=label+1; *p; p++ )
    // start of a trace index?
    if (*p == '-' || *p == '=')
      if (daughters.size() > 0 || *(p-1) == '*') {
	
	// match the following characters with the suffix
	const char *s=suffix;
	for( char *pp=p+1; ; pp++, s++ ) {
	  if (*s == 0) {
	    // matching complete
	    if (*pp == 0 || *pp == '-' || *pp == '=') {
	      char c;
	      if (*p == '-')
		c = '_';
	      else c = *p;
	      // delete the trace index
	      while ((*(p++) = *(pp++))) ;
	      return c;
	    }
	    break;
	  }
	  else if (*pp != *s)
	    break; // mismatch
	}
      }
  return 0;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::correct                                                  */
/*                                                                 */
/*FE****************************************************************/

void Tree::correct()

{
  // time markers a.m., p.m., a.m, p.m, o'clock
  if (daughters.size() == 0) {

    if (strcmp(label,"o'clock") == 0 ||
	((label[0] == 'a' || label[0] == 'p') &&
	 label[1] == '.' && label[2] == 'm' &&
	 ((label[3] == '.' && label[4] == 0) || label[3] == 0)))
      mother->change_label("NN/time");

    return;
  }

  // copy the LGS feature from NPs to the dominating PPs
  if ((matching_category("NP-LGS") || 
       matching_category("S-NOM-LGS") ||
       matching_category("SBAR-NOM-LGS")) &&
       mother->matching_category("PP"))
    {
      if (mother->mother->matching_category("PP"))
	mother->mother->insert_uniq_feature("lgs");
      else
	mother->insert_uniq_feature("lgs");
    }

  // simplify the PENN categories
  char *b,*e;
  split_catname(label, &b, &e);
  char *l = CatMap[b];
  if (l) {
    char buffer[1000];
    free(label);
    sprintf(buffer,"%s%s",l,e);
    label = strdup(buffer);
  }

  if (matching_category("RB")) {

    // correct adjectives which are tagged as adverbs
    if (mother->matching_base_cat("ADJP")) {
      for( int i=mother->daughters.size()-1; i>=0; i-- ) {
	if (mother->daughters[i] == this)
	  change_label("JJ/corr");
	else if (mother->daughters[i]->matching_base_cat("PP") ||
		 mother->daughters[i]->matching_base_cat("S") ||
		 mother->daughters[i]->matching_base_cat("SBAR") ||
		 mother->daughters[i]->matching_base_cat("NP"))
	  continue;
	break;
      }
    }
    char *l=lowercase(daughters[0]->label);
    if (strcmp(l,"tonight") == 0 || strcmp(l,"tomorrow") == 0 || 
	strcmp(l,"yesterday") == 0 || strcmp(l,"") == 0)
      change_label("NN");
  }

  else if (matching_base_cat("ADJP")) {
    // sequence of conjuncts?
    size_t i;
    for( i=1; i<daughters.size(); i+=2 )
      if (strcmp(daughters[i]->label,"CC") != 0 && 
	  strcmp(daughters[i]->label,",") != 0)
	break;
    if (i == daughters.size())
      for( i=0; i<daughters.size(); i+=2 )
	if (strcmp(daughters[i]->label,"NN") == 0 ||
	    strcmp(daughters[i]->label,"NNS") == 0 ||
	    strcmp(daughters[i]->label,"NNP") == 0)
	  daughters[i]->change_label("JJ");
  }

  // correct possessive markers which are tagged as verbs
  // and nouns which are tagged as verbs
  else if (label[0] == 'V' && label[1] == 'B' &&
	   mother->matching_base_cat("NP") && is_last_daughter())
    {
      if (strcmp(daughters[0]->label,"'s") == 0)
	change_label("POS");
      else if (daughters[0]->matching_category("VBZ"))
	change_label("NNS");
      else
	change_label("NN");
    }

#if 0
  else if (strcmp(label,":") == 0 &&
	   strcmp(daughters[0]->label, "'") == 0)
    change_label("''");
#endif


  // recursion
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->correct();


  if (strcmp(label,":") == 0 && daughters.size() == 1) {
    if (strcmp(daughters[0]->label,";") == 0)
      change_label(";");
    else if (strcmp(daughters[0]->label,"--") == 0)
      change_label("--");
    else if (strcmp(daughters[0]->label,"...") == 0)
      change_label("...");
  }

  if (matching_base_cat("NP") && daughters.size() == 1 && 
	   strcmp(daughters[0]->label,"PRP/refl") == 0)
    insert_feature("refl");

#if 0
  if (strcmp(label,"VB") == 0 && mother->matching_base_cat("PP") &&
      strcmp(daughters[0]->label,"like") == 0)
    change_label("IN");
#endif

  if (strcmp(label,"PRP") == 0) {
    char *word=lowercase(daughters[0]->label);
    if (strcmp(word,"I") == 0 ||
	strcmp(word,"he") == 0 ||
	strcmp(word,"she") == 0 ||
	strcmp(word,"we") == 0 ||
	strcmp(word,"they") == 0)
      insert_feature("nom");
    else if (strcmp(word,"me") == 0 ||
	     strcmp(word,"him") == 0 ||
	     strcmp(word,"her") == 0 ||
	     strcmp(word,"us") == 0 ||
	     strcmp(word,"them") == 0)
      insert_feature("acc");
    else if (strcmp(word,"myself") == 0 ||
	     strcmp(word,"yourself") == 0 ||
	     strcmp(word,"himself") == 0 ||
	     strcmp(word,"herself") == 0 ||
	     strcmp(word,"itself") == 0 ||
	     strcmp(word,"ourselves") == 0 ||
	     strcmp(word,"yourselves") == 0 ||
	     strcmp(word,"themselves") == 0)
      insert_feature("refl");
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  trace_category                                                 */
/*                                                                 */
/*FE****************************************************************/

static char *trace_category( char *label )

{
  static char buffer[1000];
  char *p=buffer;
  bool copy=true;
  for( ; *label; label++ ) {
    if (*label == '/')
      copy = false;
    // ignore Penn features other than NP-SBJ
    else if (*label == '-') {
      if (strcmp(buffer,"NP") == 0)
	label++;
      else
	copy = false;
    }
    else if (*label == '=' || *label == '%' || *label == '+' || 
	     index_start(label))
      break;
    else if (*label == '_')
      copy = true;
    if (copy) {
      *(p++) = *label;
      *p = 0;
      // simplify SINV and SQ to S
      if (strcmp(buffer,"SINV") == 0)
	p -= 3;
      else if (strcmp(buffer,"SQ") == 0)
	p--;
    }
  }
  *p = 0;
  return buffer;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::append_slashes                                           */
/*                                                                 */
/*FE****************************************************************/

char Tree::append_slashes( char *prefix, const char *suffix,
			   bool &terminal )
{
  // current node carries the index?
  char c;
  if ((c=matching_trace(suffix))) {
    terminal = true;
    // mark S-traces which are dominated by SBAR
    if (prefix[0] == 'S') {
      if (mother->matching_category("-NONE-") &&
	  mother->mother->label[0] == 'S' &&
	  mother->mother->label[1] != 'B' &&
	  mother->mother->mother->matching_category("SBAR") &&
	  mother->mother->mother->daughters[0]->matching_category("-NONE-"))
	prefix[0] = 'X';
    }
  }
  else {
    terminal = false;
    // look for the index in the subtree
    for( int i=(int)daughters.size()-1; i>=0; i-- ) {
      bool term;
      if ((c = daughters[i]->append_slashes(prefix, suffix, term))) {
	// match in subtree
	daughters[i]->append_slash( prefix, c, term );
	break;
      }
    }
  }

  return c;
}



/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_slashes                                              */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_slashes()

{
  if (daughters.size() == 0)
    return;

  // process daughter nodes
  for( int i=(int)daughters.size()-1; i>=0; i-- )
    daughters[i]->add_slashes();

  // Is current node a filler?
  char *s;
  while ((s = find_index(label)) != NULL) {
    *(s++) = 0; // delete the index
    char *p=trace_category(label);
    bool match_found = false;
    s = strdup(s);
    for(;;){
      Tree *t;
      char c=0;
      bool term; // not needed actually
      for( t=this; t; t=t->mother )
	if ((c = t->append_slashes( p, s, term )) != 0)
	  break;
      if (t == NULL)
	break; // exit loop

      match_found = true;
      if (t == this)
	append_slash( p, '+', true );
      else {
	if (c == '_')
	  c = '+';
	else
	  c = '%';
	append_slash( p, c, true );
	for( Tree *t2=mother; t2 != t; t2=t2->mother )
	  t2->append_slash( p, c, false );
      }
    }
    if (Verbose && !match_found)
      fprintf(stderr,"\tWarning: no matching trace for \"%s-%s\"!\n",label,s);
    free(s);
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::matching_prefix                                          */
/*                                                                 */
/*FE****************************************************************/

bool Tree::matching_prefix( const char *prefix )

{
  for( char *p=label; *prefix; p++, prefix++ )
    if (*p != *prefix)
      return false;
  return true;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::has_trace_feature                                        */
/*                                                                 */
/*FE****************************************************************/

bool Tree::has_trace_feature( char d, const char *s )

{
  for( const char *p=label; *p; p++ )
    if (*p == d) {
      const char *p1=p+1;
      if (*p1 == '#')
	p1++;
      if (*p1 == '') {
	const char *p2=s;
	p1++;
	// compare the categories
	while (*p2 && *(p1++) == *(p2++)) ;
	if (*p1 == '')
	  return true;
      }
  }
  return false;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_genitive_markers                                     */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_genitive_markers()

{
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_genitive_markers();

  if (matching_category("NP"))
    for( int i=daughters.size()-1; i>0; i-- ) {
      if (daughters[i]->matching_category("POS"))
	insert_feature("gen");
      else if (daughters[i]->matching_category("''"))
	continue;
      break;
    }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_TMP_markers                                          */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_TMP_markers()

{
  // mark NP daughter of PP-TMP as tmp
  if (matching_base_cat("PP-TMP")) {
    int i=0;
    while (i < (int)daughters.size() && 
	   !daughters[i]->matching_category("IN"))
      { i++; }
    for( ; i<(int)daughters.size(); i++)
      if (daughters[i]->matching_category("NP")) {
	daughters[i]->insert_feature("tmp");
	break;
      }
  }

  // mark NP daughter of NP-TMP as tmp
  else if (matching_base_cat("NP-TMP")) {
    for( size_t i=0; i<daughters.size(); i++ )
#if 0
      if (daughters[i]->matching_category("NP")) {
	daughters[i]->insert_feature("tmp");
	break;
      }
      else
#endif
      if (daughters[i]->matching_category("JJ")) {
	char *s=lowercase(daughters[i]->daughters[0]->label);
	if (strcmp(s,"last") == 0 || strcmp(s,"next") == 0 ||
	    strcmp(s,"late") == 0 || strcmp(s,"previous") == 0 ||
	    strcmp(s,"early") == 0|| strcmp(s,"past") == 0)
	  {
	    daughters[i]->insert_feature("tmp");
	    break;
	  }
      }
  }

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_TMP_markers();

  if (matching_base_cat("NP-TMP") || matching_base_cat("NP/tmp")) {
    for( int i=(int)daughters.size()-1; i>=0; i-- )
      if (daughters[i]->noun_POS_tag())	{
	daughters[i]->insert_feature("tmp");

#if 0
	// tomorrow morning
	if (i > 0 && daughters[i-1]->matching_category("NN")) {
	  char *s=lowercase(daughters[i-1]->daughters[0]->label);
	  if (strcmp(s,"today") == 0 ||
	      strcmp(s,"tomorrow") == 0 ||
	      strcmp(s,"yesterday") == 0)
	    daughters[i-1]->insert_feature("tmp2");
	}
#endif
	    
	while (i >= 2 &&
	       daughters[--i]->matching_category("CC") &&
	       daughters[--i]->noun_POS_tag())
	  {
	    daughters[i]->insert_feature("tmp");
	  }
	return;
      }
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_MNR_markers                                          */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_MNR_markers()

{
  if (matching_base_cat("NP-MNR") &&
      daughters.size() == 2 &&
      daughters[0]->matching_category("NP") &&
      daughters[1]->matching_category("SBAR"))
    daughters[0]->insert_feature("mnr");

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_MNR_markers();

  if (matching_base_cat("NP-MNR") || matching_base_cat("NP/mnr")) {
    for( int i=(int)daughters.size()-1; i>=0; i-- )
      if (daughters[i]->matching_category("NN") ||
	  daughters[i]->matching_category("NNS"))
	{
	  daughters[i]->insert_feature("mnr");
	  return;
	}
  }

  else if (matching_base_cat("ADVP-MNR")) {
    for( int i=(int)daughters.size()-1; i>=0; i-- ) {
      if (daughters[i]->matching_category("RB") ||
	  daughters[i]->matching_category("RB?"))
	{
	  daughters[i]->insert_feature("mnr");
	  return;
	}
      else if (daughters[i]->matching_base_cat("PP") ||
	       daughters[i]->matching_base_cat("NP") ||
	       daughters[i]->matching_base_cat("S") ||
	       daughters[i]->matching_base_cat("SBAR") ||
	       daughters[i]->matching_base_cat(","))
	continue;
      break;
    }
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_NPqp_markers                                         */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_NPqp_markers()

{
  if (matching_base_cat("NP-EXT")) {
    int i;
    for( i=(int)daughters.size()-1; i>=0; i-- )
      if (daughters[i]->noun_POS_tag())	{
	daughters[i]->insert_feature("ext");
	break;
      }
    if (i < 0)
      for( size_t i=0; i<daughters.size()-1; i++ )
	if (daughters[i]->matching_category("NP"))
	  daughters[i]->change_label("NP-EXT");
  }

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_NPqp_markers();

  // (NP/qp (NP/qp ...) ... (NP-ADV ...) ...)
  if (matching_base_cat("NP-ADV")) {
    insert_feature(mother->base_cat());
    if (mother->daughters[0] != this &&
	mother->daughters[0]->matching_category("NP"))
      {
	if (mother->matching_base_cat("NP"))
	  mother->insert_uniq_feature("qp");
	mother->daughters[0]->insert_uniq_feature("qp");
      }
  }

  // currency and percent markers
  else if (matching_category("NP") || matching_category("QP")) {
    for( size_t i=0; i<daughters.size(); i++ )
      if (strcmp(daughters[i]->label, "$") == 0 ||
	  strcmp(daughters[i]->label, "#") == 0 ||
	  strcmp(daughters[i]->label, "SYM") == 0)
	insert_uniq_feature("$");
      else if (strcmp(daughters[i]->label, "NN/%") == 0)
	insert_uniq_feature("%");
  }

  else if (matching_category("NN") &&
           strcmp(daughters[0]->label, "%") == 0)
    insert_uniq_feature("%");

  else if (daughters.size() == 0 && strcmp(label, "%") == 0)
    mother->insert_uniq_feature("%");
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_PP_features                                          */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_PP_features()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_PP_features();

  if (matching_base_cat("PP"))
    insert_feature(mother->base_cat());

  if (matching_category("IN") && daughters.size() == 1) {
    char *l=lowercase(daughters[0]->label);
    if (PrepMap.find(l) != PrepMap.end())
      insert_feature(PrepMap[l]);
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_ADV_features                                         */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_ADV_features()

{
  if (daughters.size() == 0)
    return;
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_ADV_features();

  if (matching_category("RB") || matching_category("RB?")) {

    // mark adverbs with irregular behaviour
    if (strcmp(lowercase(daughters[0]->label), "as") == 0)
      insert_feature("as");
    else if (strcmp(lowercase(daughters[0]->label), "so") == 0)
      insert_feature("so");
    else if (strcmp(lowercase(daughters[0]->label), "about") == 0)
      insert_feature("about");
    else if (strcmp(lowercase(daughters[0]->label), "not") == 0 ||
	     strcmp(daughters[0]->label, "n't") == 0)
      insert_feature("not");

    // mark adverbs with parent information
    else if (mother->matching_base_cat("ADVP") &&
	     mother->daughters.size() == 1)
      insert_feature(mother->mother->base_cat());

    else
      insert_feature(mother->base_cat());
  }
  else if (matching_base_cat("ADVP")) {
#if 0
    if (mother->matching_base_cat("PP") &&
	predecessor() != NULL &&
	predecessor()->matching_category("IN"))
      insert_feature("pparg");
    else
#endif
    if (mother->matching_base_cat("PP") &&
	     is_last_daughter())
      insert_feature("ppfinal");
    else
      insert_feature(mother->base_cat());

    if (daughters.size() > 1) {
      bool mwl=true;
      Tree *head=NULL;
      for( size_t i=0; i<daughters.size(); i++ ) {
	// add subcat feature
	if (head != NULL) {
	  if (daughters[i]->matching_category("PP"))
	    head->insert_feature("p");
	  else if (daughters[i]->matching_category("NP"))
	    head->insert_feature("n");
	  else if (daughters[i]->matching_category("S"))
	    head->insert_feature("s");
	  else if (daughters[i]->matching_category("SBAR"))
	    head->insert_feature("b");
	}

	if (daughters[i]->matching_category("RB") ||
	    daughters[i]->matching_category("RB?") ||
	    daughters[i]->matching_category("ADVP"))
	  {
	    head = daughters[i];
	    mwl = false;
	    if (i == daughters.size()-1 && i > 0) {
	      if (daughters[i-1]->matching_base_cat("NP"))
		daughters[i]->insert_feature("m");
	    }
	  }
	else if (daughters[i]->daughters.size() != 1 ||
		 daughters[i]->daughters[0]->daughters.size() > 0)
	  {
	    mwl = false;
	  }
      }
      if (mwl) {
	for( size_t i=0; i<daughters.size(); i++ ) {
	  char *l=lowercase(daughters[i]->daughters[0]->label);
	  if (strcmp(l,"least") == 0 || strcmp(l,"all") == 0 ||
	      strcmp(l,"and") == 0 || strcmp(l,"more") == 0 ||
	      strcmp(l,"no") == 0 || strcmp(l,"sort") == 0 ||
	      strcmp(l,"any") == 0 || strcmp(l,"kind") == 0 ||
	      strcmp(l,"longer") == 0 || strcmp(l,"long") == 0 ||
	      strcmp(l,"door") == 0 || strcmp(l,"best") == 0 ||
	      strcmp(l,"way") == 0 || strcmp(l,"right") == 0 ||
	      strcmp(l,"last") == 0 || strcmp(l,"first") == 0 ||
	      strcmp(l,"next") == 0 || strcmp(l,"neck") == 0 ||
	      strcmp(l,"large") == 0 || strcmp(l,"home") == 0 ||
	      strcmp(l,"head") == 0 || strcmp(l,"doubt") == 0 ||
	      strcmp(l,"time") == 0 || strcmp(l,"same") == 0 ||
	      strcmp(l,"part") == 0 || strcmp(l,"or") == 0 ||
	      strcmp(l,"much") == 0 || strcmp(l,"course") == 0 ||
	      strcmp(l,"back") == 0)
	    daughters[i]->insert_feature(l);
	}
      }
    }
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_ADJ_features                                         */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_ADJ_features()

{
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_ADJ_features();

  if (matching_base_cat("ADJP")) {

    // mark attributive and post-modifying adjectives
    if (mother->matching_base_cat("NP")) {
      if (is_last_daughter())
	insert_feature("post");
      else
	insert_feature("attr");
    }

    for( size_t i=0; i<daughters.size(); i++ ) 
      if (daughters[i]->matching_category("JJ")) {

	// mark predicative adjectives
	if (matching_base_cat("ADJP-PRD"))
	  daughters[i]->insert_feature("prd");

	// add subcat feature
	if (i < daughters.size()-1) {
	  if (daughters[i+1]->matching_category("PP"))
	    daughters[i]->insert_feature("p");
	  else if (daughters[i+1]->matching_category("NP"))
	    daughters[i]->insert_feature("n");
	  else if (daughters[i+1]->matching_category("S"))
	    daughters[i]->insert_feature("s");
	  else if (daughters[i+1]->matching_category("SBAR"))
	    daughters[i]->insert_feature("b");
	}
	else if (i > 0) {
	  if (daughters[i-1]->matching_base_cat("NP"))
	    daughters[i]->insert_feature("m");
	}
      }
#if 0
      else if (daughters[i]->matching_category("NN") ||
	       daughters[i]->matching_category("NNS")) {
	daughters[i]->insert_feature("adjp");
      }
#endif
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_DT_features                                          */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_DT_features()

{
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_DT_features();
    
  if (daughters.size() > 0) {
    char *word=lowercase(daughters[0]->label);

    if (label[0] == 'W') {
      if (strcmp(word,"that") == 0 || strcmp(word,"which") == 0 ||
	  strcmp(word,"what") == 0 || strcmp(word,"who") == 0 ||
	  strcmp(word,"how") == 0)
	insert_feature(word);
    }
    else if (matching_category("DT")) {
      if (strcmp(word,"a") == 0 || strcmp(word,"an") == 0)
	insert_feature("a");
      else if (strcmp(word,"the") == 0)
	insert_feature("the");
      else if (strcmp(word,"this") == 0  || strcmp(word,"that") == 0 ||
	       strcmp(word,"these") == 0 || strcmp(word,"those") == 0)
	insert_feature("this");
    }
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_NP_features                                          */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_NP_features()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_NP_features();

  if (matching_base_cat("NP")) {
    // (NP (NP ...) (SBAR (WHADVP (-NONE- 0)) ...))
    Tree *sbar=successor();
    if (sbar && sbar->matching_category("SBAR") &&
	mother->matching_base_cat("NP") &&
	sbar->daughters[0]->matching_category("WHADVP") &&
	sbar->daughters[0]->daughters[0]->matching_category("-NONE-") &&
	strcmp(sbar->daughters[0]->daughters[0]->daughters[0]->label,"0") == 0)
      {
	Tree *d=last_daughter();
	if (d->matching_category("NN") || d->matching_category("NNS")) {
	  char *word=d->daughters[0]->label;
	  if (strcmp(word,"way") == 0 || strcmp(word,"ways") == 0 || 
	      strcmp(word,"reason") == 0 || strcmp(word,"reasons") == 0 || 
	      strcmp(word,"day") == 0 || strcmp(word,"days") == 0 || 
	      strcmp(word,"time") == 0 || strcmp(word,"moment") == 0 || 
	      strcmp(word,"place") == 0 || strcmp(word,"position") == 0)
	    {
	      d->insert_feature("whadvp");
	      insert_feature("whadvp");
	      sbar->insert_feature("whadvp");
	    }
	}
      }

    // add verbal agreement feature
    if (matching_category("NP-SBJ")) {
      const char *f=NULL;
      for( size_t i=0; i<mother->daughters.size(); i++ )
	if (mother->daughters[i]->matching_category("VP/3s")) {
	  f = "3s";
	  break;
	}
	else if (mother->daughters[i]->matching_category("VP/n3s")) {
	  f = "n3s";
	  break;
	}
      if (f) {
	insert_feature(f);
	for( size_t i=0; i<daughters.size(); i++ )
	  if (daughters[i]->matching_category("NP") ||
	      daughters[i]->matching_category("PRP"))
	    {
	      daughters[i]->insert_feature(f);
	      break;
	    }
      }
      
      // mark constructions of the form
      // (S (NP-SBJ (NP (PRP It))
      //            (S (-NONE- *EXP*-1)))
      //    (VP (VBZ is)
      //        (ADJP-PRD (JJ important))
      //        (S-1 (NP-SBJ (-NONE- *))
      //             (VP (TO to)
      //                 (VP (VB realize)...)))))
      if (daughters.size() == 2 && 
	  daughters[0]->matching_category("NP") &&
	  daughters[0]->daughters[0]->matching_category("PRP") &&
	  daughters[1]->matching_category("S") &&
	  daughters[1]->daughters[0]->matching_category("-NONE-"))
	{
	  daughters[0]->insert_feature("expl");
	  daughters[0]->daughters[0]->insert_feature("expl");
	}
    }
    
    for( int i=daughters.size()-1; i>=0; i-- ) {
      Tree *d=daughters[i];
      
      // add subcat features
      if (d->matching_category("NN") || d->matching_category("NNS")) {
	for( size_t k=i+1; k<daughters.size(); k++ ) {
	  if (daughters[k]->matching_category("RB"))
	    continue;
	  if (daughters[k]->matching_base_cat("PP") &&
	      (!daughters[k]->daughters[0]->matching_category("IN") ||
	       strcmp(daughters[k]->daughters[0]->daughters[0]->label,"of")))
	    d->insert_feature("p");
	  else if (daughters[k]->matching_base_cat("S"))
	    d->insert_feature("s");
	  else if (daughters[k]->matching_base_cat("SBAR"))
	    d->insert_feature("b");
	  break;
	}
	break;
      }
      
      // add feature if modified by a PP
      else if (d->matching_base_cat("PP")) {
	insert_uniq_feature("pp");
	if (strcmp(d->daughters[0]->daughters[0]->label,"of") == 0)
	  d->insert_feature("of");
      }
    }
    
    // add base noun chunk feature
    for( size_t i=0; i<daughters.size(); i++ )
      if (daughters[i]->noun_POS_tag() ||
	  daughters[i]->matching_category("DT") ||
	  daughters[i]->matching_category("CD") ||
	  daughters[i]->matching_category("JJ") ||
	  daughters[i]->matching_category("JJ?") ||
	  daughters[i]->matching_category("PRP") ||
	  daughters[i]->matching_category("RB") ||
	  daughters[i]->matching_category("EX"))
	{
	  insert_feature("base");
	  return;
	}

#if 0
    // (NP (NP ...) (NP (DT both|all|each)))
    if (daughters.size() == 2 &&
	daughters[0]->matching_category("NP") &&
	daughters[1]->matching_category("NP") &&
	daughters[1]->daughters.size() == 1 &&
	daughters[1]->daughters[0]->matching_category("DT"))
      {
	daughters[1]->insert_feature("postDT");
	daughters[1]->daughters[0]->insert_feature("postDT");
      }
#endif
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::VP_error                                                 */
/*                                                                 */
/*FE****************************************************************/

void Tree::VP_error()

{
  fputs("\tWarning: problem in VP rule:", stderr);
  for( size_t i=0; i<daughters.size(); i++ )
    fprintf(stderr, " %s", daughters[i]->label);
  fputc('\n', stderr);
}	


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_CD_features                                          */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_CD_features()

{
  if (daughters.size() == 1 && matching_category("CD") &&
      is_year_number(daughters[0]->label))
    insert_feature("year");

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_CD_features();
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_V_features                                           */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_V_features()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_V_features();

  if (matching_category("VB") || matching_category("VB?")) {
    label[1] = verb_type(daughters[0]->label);
    if (mother->matching_base_cat("PP"))
      insert_feature("prep");
  }

  else if (matching_base_cat("VP") || matching_base_cat("SQ")) {

    // look for the verbal head
    for( size_t i=0; i<daughters.size(); i++ )
      if (daughters[i]->matching_category("V?") ||
	  daughters[i]->matching_category("V??") ||
	  daughters[i]->matching_category("-NONE-") ||
	  daughters[i]->matching_category("MD") ||
	  daughters[i]->matching_category("TO") ||
	  daughters[i]->matching_category("FRAG"))
	return;

      else {
	// correct tagging errors
	if (daughters[i]->matching_category("NNS")) {
	  daughters[i]->change_label("VVZ");
	  return;
	}
	if (daughters[i]->matching_category("POS")) {
	  daughters[i]->change_label("VBZ");
	  return;
	}
	if (daughters[i]->matching_category("IN")) {
	  daughters[i]->change_label("VV");
	  return;
	}
      }
    
    if (matching_base_cat("VP")) {
      size_t i;
      for( i=0; i<daughters.size(); i++ )
	if (strchr(daughters[i]->label, '='))
	  break;
      if (Verbose && i == daughters.size())
	VP_error();
    }
  }
}



/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_subcat_feature                                       */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_subcat_feature()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_subcat_feature();

  if (matching_base_cat("VP") || matching_base_cat("SQ") || 
      matching_base_cat("SINV")) {

    for( size_t i=0; i<daughters.size(); i++ )
      if (daughters[i]->matching_category("VP"))
	return;
      else if (daughters[i]->matching_category("V?") || 
	       daughters[i]->matching_category("V??"))
	{
	  char features[1000], *p=features;
	  *p = '0';
	  *(p+1) = 0;
	  for( size_t k=i+1; k<daughters.size(); k++ ) {
	    if (daughters[k]->matching_category("NP/refl"))
	      append_string(p, "r");
	    else if (daughters[k]->matching_category("NP") ||
		daughters[k]->matching_base_cat("SBAR-NOM") ||
		daughters[k]->matching_base_cat("NP-PRD"))
	      append_string(p, "n");
	    else if (daughters[k]->matching_base_cat("ADJP-PRD"))
	      append_string(p, "j");
 	    else if (daughters[k]->matching_base_cat("ADVP-PRD"))
 	      append_string(p, "a");
	    else if (daughters[k]->matching_base_cat("PRT"))
	      append_string(p, "t");
	    else if (daughters[k]->matching_category("PP") ||
		     daughters[k]->matching_category("PP-CLR") ||
		     daughters[k]->matching_category("PP-DTV") ||
		     daughters[k]->matching_category("PP-DIR"))
	      append_string(p, "p");
	    else if (daughters[k]->matching_category("SBAR") ||
		     daughters[k]->matching_category("SBAR-CLR"))
	      append_string(p, "b");
	    else if (daughters[k]->matching_category("S/fin"))
	      append_string(p, "sf");
 	    else if (daughters[k]->matching_category("S/ppres/gap") ||
		     daughters[k]->matching_category("S-PRD/ppres/gap") ||
		     daughters[k]->matching_category("S-CLR/ppres/gap"))
 	      append_string(p, "sg");
 	    else if (daughters[k]->matching_category("S/to/gap") ||
		     daughters[k]->matching_category("S-PRD/to/gap") ||
		     daughters[k]->matching_category("S-CLR/to/gap"))
 	      append_string(p, "st");
	    else if (daughters[k]->matching_category("S") ||
		     daughters[k]->matching_category("S-CLR") ||
		     daughters[k]->matching_category("S-PRD"))
	      append_string(p, "so");
	    else if (daughters[k]->matching_category("VP/ppres"))
	      append_string(p, "vg");
	    else if (daughters[k]->matching_category("VP/ppast"))
	      append_string(p, "vn");
	    else if (daughters[k]->matching_category("VP/pas"))
	      append_string(p, "vp");
	    else if (daughters[k]->matching_category("VP/inf"))
	      append_string(p, "vi");
	    else if (daughters[k]->matching_category("VP"))
	      append_string(p, "vo");
	  }
	  daughters[i]->insert_feature(features);
	  return;
	}
      else if (daughters[i]->matching_category("MD")) {
#if 0
	for( size_t k=i+1; k<daughters.size(); k++ ) {
	  if (daughters[k]->matching_category("VP/to"))
	    daughters[i]->insert_feature("to");
	  else if (daughters[k]->matching_category("VP/ppres"))
	    daughters[i]->insert_feature("ing");
	  else if (daughters[k]->matching_category("VP/ppast"))
	    daughters[i]->insert_feature("ed");
	  else if (daughters[k]->matching_category("VP/pas"))
	    daughters[i]->insert_feature("pas");
	  else if (daughters[k]->matching_category("VP/inf"))
	    daughters[i]->insert_feature("inf");
	  else if (daughters[k]->matching_category("S/to"))
	    daughters[i]->insert_feature("to");
	  else if (daughters[k]->matching_category("VP") &&
		   daughters[k]->is_trace())
	      ;
	  else
	    continue;
	  break;
	}
#endif
	return;
      }
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_punctuation_feature                                  */
/*                                                                 */
/*FE****************************************************************/

char *Tree::add_punctuation_feature()

{
  char *p=NULL;
  if (daughters.size() > 0) {
    for( int i=(int)daughters.size()-1; i>=0; i-- )
      if ((p = daughters[i]->add_punctuation_feature()) != NULL)
	break;

    if (p && mother)
      insert_feature(p);

    if (daughters.size() == 1)
      if (strcmp(label,".") == 0 ||
	  strcmp(label,"!") == 0 ||
	  strcmp(label,"?") == 0 ||
	  strcmp(label,":") == 0)
	p = label;
  }
  return p;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_VP_features                                          */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_VP_features()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_VP_features();

  if (matching_category("VP")) {
    size_t i;
    for( i=0; i<daughters.size(); i++ ) {
      if (daughters[i]->matching_category("TO"))
	insert_feature("to");
      else if (daughters[i]->matching_category("V?G"))
	insert_feature("ppres");
      else if (daughters[i]->matching_category("V?N")) {
	bool nptrace=false;
	for( unsigned k=i+1; k<daughters.size(); k++ )
	  if (daughters[k]->matching_category("NP") &&
	      daughters[k]->daughters[0]->matching_category("-NONE-"))
	    {
	      nptrace = true;
	      break;
	    }
	if (nptrace)
	  insert_feature("pas");
	else
	  insert_feature("ppast");
      }
      else if (daughters[i]->matching_category("V?Z")) {
	insert_feature("3s");
      }
      else if (daughters[i]->matching_category("VBD") &&
 	       strcmp(daughters[i]->daughters[0]->label,"was")==0)
 	{
 	  insert_feature("3s");
	  // daughters[i]->insert_feature("3s");
 	}
      else if (daughters[i]->matching_category("V?P")) {
	insert_feature("n3s");
      }
      else if (daughters[i]->matching_category("VBD") &&
 	       strcmp(daughters[i]->daughters[0]->label,"were")==0)
 	{
 	  insert_feature("n3s");
	  // daughters[i]->insert_feature("n3s");
 	}
      else if (daughters[i]->matching_category("V?D") ||
	       daughters[i]->matching_category("MD")) {
	insert_feature("fin");
      }
      else if (daughters[i]->matching_category("V?") &&
	       !daughters[i]->matching_category("VP"))
	insert_feature("inf");
      else if (daughters[i]->matching_category("VP")) {
	if (daughters[i]->matching_category("VP/to"))
	  insert_feature("to");
	else if (daughters[i]->matching_category("VP/ppres"))
	  insert_feature("ppres");
	else if (daughters[i]->matching_category("VP/ppast"))
	  insert_feature("ppast");
	else if (daughters[i]->matching_category("VP/pas"))
	  insert_feature("pas");
	else if (daughters[i]->matching_category("VP/inf"))
	  insert_feature("inf");
	else if (daughters[i]->matching_category("VP/fin"))
	  insert_feature("fin");
	else if (daughters[i]->matching_category("VP/3s"))
	  insert_feature("3s");
	else if (daughters[i]->matching_category("VP/n3s"))
	  insert_feature("n3s");
	else
	  continue;
      }
      else
	continue;
      break;
    }
    if (i == daughters.size()) {
      if (mother->daughters[0]->matching_category("TO") &&
	  daughters[0]->matching_category("NN"))
	{
	  daughters[0]->change_label("VV");
	  insert_feature("inf");
	}
    }
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_S_features                                           */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_S_features()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_S_features();

  if (daughters.size() == 0)
    return;
  
  if (matching_base_cat("S") && !matching_category("S/imp")) {
    size_t k;
    for( k=0; k<daughters.size(); k++)
      if (daughters[k]->matching_category("VP"))
	break;
    if (k < daughters.size()) {
      if (daughters[k]->matching_category("VP/to"))
	insert_feature("to");
      else if (daughters[k]->matching_category("VP/fin") ||
	       daughters[k]->matching_category("VP/3s") ||
	       daughters[k]->matching_category("VP/n3s"))
	{
	  if ((mother->matching_category("TOP") ||
	       mother->matching_category("S-ADV")) &&
	      mother->daughters[0]->matching_category("NP-SBJ") &&
	      mother->daughters[0]->is_trace())
	    insert_feature("imp");
	  else
	    insert_feature("fin");
	}
      else if (daughters[k]->matching_category("VP/ppast"))
	insert_feature("ppast");
      else if (daughters[k]->matching_category("VP/pas"))
	insert_feature("pas");
      else if (daughters[k]->matching_category("VP/ppres"))
	insert_feature("ppres");
      else if (daughters[k]->matching_category("VP/inf")) {
	for( size_t l=0; l<k; l++)
	  if (daughters[l]->matching_category("NP-SBJ") && 
	      daughters[l]->is_trace())
	    {
	      insert_feature("imp");
	      break;
	    }
      }
      else {
	for( size_t l=0; l<k; l++)
	  if (daughters[l]->matching_category("NP-SBJ")) {
	    insert_feature("fin");
	    daughters[k]->insert_feature("fin");
	    if (daughters[k]->daughters[0]->matching_category("NN") ||
		daughters[k]->daughters[0]->matching_category("JJ"))
	      daughters[k]->daughters[0]->change_label("VVP");
	    break;
	  }
      }
    }

    else { // no VP daughter
      int k=daughters.size();
      while (--k >= 0)
	if (daughters[k]->matching_category("S"))
	  break;
      while (--k >= 0)
	if (daughters[k]->matching_category("CC") ||
	    daughters[k]->matching_category("CONJP") ||
	    daughters[k]->matching_category(",") ||
	    daughters[k]->matching_category(":"))
	  break;
      while (--k >= 0)
	if (daughters[k]->matching_category("S"))
	  break;
      if (k >= 0) {
	if (daughters[k]->matching_base_cat("S/fin"))
	  insert_feature("fin");
	else if (daughters[k]->matching_base_cat("S/imp"))
	  insert_feature("imp");
	else if (daughters[k]->matching_base_cat("S/ppres"))
	  insert_feature("ppres");
	else if (daughters[k]->matching_base_cat("S/ppast"))
	  insert_feature("ppast");
	else if (daughters[k]->matching_base_cat("S/pas"))
	  insert_feature("pas");
	else if (daughters[k]->matching_base_cat("S/to"))
	  insert_feature("to");
      }
      else {
	size_t k;
	for( k=0; k<daughters.size(); k++)
	  if (daughters[k]->matching_category("NP-SBJ") ||
	      daughters[k]->matching_category("S-NOM") ||
	      daughters[k]->matching_category("SBAR-NOM") ||
	      daughters[k]->matching_category("PP-NOM"))
	    break;
	  else if (daughters[k]->matching_category("S/fin")) {
	    insert_feature("fin");
	    return;
	  }
	for( ; k<daughters.size(); k++) {
	  if (daughters[k]->matching_base_cat("ADJP-PRD"))
	    insert_feature("adjp");
	  else if (daughters[k]->matching_base_cat("ADVP-PRD"))
	    insert_feature("advp");
	  else if (daughters[k]->matching_base_cat("NP") &&
		   !daughters[k]->matching_base_cat("NP-SBJ"))
	    insert_feature("np");
	  else if (daughters[k]->matching_category("S-NOM"))
	    insert_feature("s");
	  else if (daughters[k]->matching_category("VP-NOM"))
	    insert_feature("v");
	  else if (daughters[k]->matching_category("PP-PRD"))
	    insert_feature("p");
	  else
	    continue;
	  break;
	}
      }
    }
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_CC_features                                          */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_CC_features()

{
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_CC_features();

  if (strcmp(label,"CC") == 0) {
    if (strcmp(daughters[0]->label, "but") == 0 ||
	strcmp(daughters[0]->label, "But") == 0)
      insert_feature("but");
    else if (strcmp(daughters[0]->label, "&") == 0)
      insert_feature("&");
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_UCP_features                                         */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_UCP_features()

{
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_UCP_features();

  if (matching_base_cat("UCP")) {
    size_t  i;
    for( i=0; i<daughters.size(); i++ ) {
      if (daughters[i]->matching_base_cat("ADJP") ||
	  daughters[i]->matching_category("JJ") ||
	  daughters[i]->matching_category("NN") ||
	  daughters[i]->matching_category("NNP"))
	insert_feature("ADJP");
      else if (daughters[i]->matching_base_cat("ADVP"))
	insert_feature("ADVP");
      else if (daughters[i]->matching_base_cat("PP"))
	insert_feature("PP");
      else if (daughters[i]->matching_base_cat("SBARQ"))
	insert_feature("SBARQ");
      else
	continue;
      return;
    }

    for( i=0; i<daughters.size(); i++ )
      if (daughters[i]->matching_base_cat("NP")) {
	insert_feature("NP");
	break;
      }
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_LST_features                                         */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_LST_features()

{
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_LST_features();

  if (matching_category("LST"))
    mother->insert_feature("lst");
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_CONJP_features                                       */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_CONJP_features()

{
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_CONJP_features();

  if (matching_category("CONJP"))
    for( size_t i=0; i<daughters.size(); i++ ) {
      daughters[i]->strip_features();
      daughters[i]->insert_feature(lowercase(daughters[i]->daughters[0]->label));
    }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_SBAR_features                                        */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_SBAR_features()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_SBAR_features();


  if (matching_base_cat("SBAR")) {
    if (matching_category("SBAR")) {
      insert_feature(mother->base_cat());
      if (mother->matching_base_cat("NP") && is_first_daughter())
	insert_feature("initial");
    }

    // (SBAR IN NN/sbar ...)
    if (daughters.size() > 2 &&
	daughters[0]->matching_category("IN") &&
	daughters[1]->matching_category("NN"))
      daughters[1]->insert_feature("sbar");
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_oc_features                                          */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_oc_features()

{
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_oc_features();

  // (VP ... (V??oc ...) ... (NP..+NP ...) ... (S.._NP ...) ...)
  if (matching_category("VP"))
    for( size_t v=0; v<daughters.size(); v++ )
      if (daughters[v]->verb_POS_tag())
	for( size_t np=v+1; np<daughters.size(); np++ )
	  if (daughters[np]->matching_category("NP") &&
	      daughters[np]->has_trace_feature('+',"NP"))
	    for( size_t s=np+1; s<daughters.size(); s++ )
	      if (daughters[s]->matching_category("S") &&
		  daughters[s]->has_trace_feature('_',"NP"))
		{
		  char buffer[100];
		  sprintf(buffer,"%soc",daughters[v]->label);
		  daughters[v]->change_label(buffer);
		  break;
		}
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_domV_feature                                         */
/*                                                                 */
/*FE****************************************************************/

bool Tree::add_domV_feature()

{
  bool result=false;
  for( size_t i=0; i<daughters.size(); i++ )
    if (daughters[i]->add_domV_feature())
      result = true;

  if (matching_base_cat("VP") || matching_base_cat("SINV") || 
      matching_base_cat("S") || matching_base_cat("SQ") || 
      matching_base_cat("SBAR") || matching_base_cat("SBARQ"))
    return true;

  if (mother && result)
    insert_feature("domV");

  return result;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::add_gapped_feature                                       */
/*                                                                 */
/*FE****************************************************************/

void Tree::add_gapped_feature()

{
  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->add_gapped_feature();

  if (matching_base_cat("S")) {
  for( size_t i=0; i<daughters.size(); i++ )
    if (daughters[i]->matching_base_cat("NP") &&
	daughters[i]->daughters[0]->matching_category("-NONE-"))
      {
	insert_feature("gap");
      }
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::mark_as_trace                                            */
/*                                                                 */
/*FE****************************************************************/

void Tree::mark_as_trace()

{
  char *buffer = (char*)malloc(strlen(label) + 3);
  sprintf(buffer, "(%s)", label);
  free(label);
  label = buffer;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::collapse_traces                                          */
/*                                                                 */
/*FE****************************************************************/

bool Tree::collapse_traces()

{
  char buffer[10000];
  char *p=buffer;

  if (daughters.size() == 0)
    return false;

  p += sprintf(p,"%s",label);
  if (matching_category("-NONE-"))
    p += sprintf(p,"(%s)",daughters[0]->label);
  else {
    bool is_trace=true;
    for( size_t i=0; i<daughters.size(); i++ )
      if (daughters[i]->collapse_traces()) {
	p += sprintf(p,"(%s)",daughters[i]->label);
	daughters[i]->mark_as_trace();
      }
      else
	is_trace = false;
    if (!is_trace)
      return false;
  }

  change_label(buffer);
  for( size_t i=0; i<daughters.size(); i++ )
    delete daughters[i];
  daughters.clear();

  return true;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::delete_indices                                           */
/*                                                                 */
/*FE****************************************************************/

void Tree::delete_indices()

{
  if (daughters.size() > 0 ||
      (mother && mother->matching_category("-NONE-")))
    {
      cut_off_indices();
      for( size_t i=0; i<daughters.size(); i++ )
	daughters[i]->delete_indices();
    }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::delete_traces                                            */
/*                                                                 */
/*FE****************************************************************/

bool Tree::delete_traces()

{
  if (daughters.size() == 0)
    return is_trace();

  bool is_trace = true;
  size_t k=0;
  for( size_t i=0; i<daughters.size(); i++ )
    if (daughters[i]->delete_traces())
      delete daughters[i];
    else {
      daughters[k++] = daughters[i];
      is_trace = false;
    }

  daughters.resize(k);
  return is_trace;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::print_govs                                               */
/*                                                                 */
/*FE****************************************************************/

void Tree::print_govs()

{
  if (daughters.size() == 0)
    return;

  for( size_t i=0; i<daughters.size(); i++ )
    daughters[i]->print_govs();

  if (mother && head != mother->head) {
    if (head)
      fprintf(govfile, "%s\t%s\t", head->daughters[0]->label, head->label);
    else
      fprintf(govfile, "<NONE>\t-NONE-\t");

    if (mother->head)
      fprintf(govfile, "%s\t%s\t", mother->head->daughters[0]->label, 
	      mother->head->label);
    else
      fprintf(govfile, "<NONE>\t-NONE-\t");
    fprintf(govfile, "%s\n", mother->label);
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::aux_label                                                */
/*                                                                 */
/*FE****************************************************************/

static bool aux_label( char *s )

{
  if (*(s++) != '<')
    return false;

  while (*s != '[') {
    if (!*s)
      return false;
    s++;
  }

  while (*s != ']') {
    if (!*s)
      return false;
    s++;
  }

  while (*s != '>') {
    if (!*s)
      return false;
    s++;
  }

  return *(s+1) == 0;
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::undo_markov                                              */
/*                                                                 */
/*FE****************************************************************/

void Tree::undo_markov()

{
  if (daughters.size() > 0) {
    Trees d;
    
    for( size_t i=0; i<daughters.size(); i++ ) {
      Tree *daughter=daughters[i];
      daughter->undo_markov();
      if (aux_label(daughter->label)) {
	for( size_t k=0; k<daughter->daughters.size(); k++ )
	  d.push_back(daughter->daughters[k]);
	daughter->daughters.clear();
	delete daughter;
      }
      else
	d.push_back(daughter);
    }
    daughters = d;
  }
}


/*FA****************************************************************/
/*                                                                 */
/*  Tree::process                                                  */
/*                                                                 */
/*FE****************************************************************/

void Tree::process()

{
  static unsigned long N=0;
  if (Verbose)
    fprintf(stderr,"\r%lu", ++N);

  if (ConvertTraces || EmptyCategories) {
    if (UndoMarkov)
      undo_markov();
    Trace_Number = 0;
    bool remaining_traces;
    do {
      remaining_traces = false;
    } while (convert(remaining_traces) && remaining_traces);

    if (remaining_traces)
      fprintf(stderr,"Warning: Some slash feature could not be translated to a trace index!\n");

    if (EmptyCategories)
      print_ec();
    else
      print();
    return;
  }

  if (UndoMarkov) {
    undo_markov();
    print();
    return;
  }

  if (WithHeads)
    assign_heads();
  correct(); // <<<
  add_genitive_markers();
  add_TMP_markers();
  add_MNR_markers();
  add_NPqp_markers();
  add_SBAR_features();
  add_V_features();
  add_CD_features();
  add_VP_features();
  add_S_features();
  add_gapped_feature();
  add_subcat_feature();
  add_NP_features();
  add_ADV_features();
  add_ADJ_features();
  add_PP_features();
  add_DT_features();
  add_CC_features();
  add_UCP_features();
  add_LST_features();
  add_CONJP_features();
  add_domV_feature();
  add_punctuation_feature();
  if (PrintTraces) {
    add_slashes();
    add_oc_features();
    collapse_traces();
  }
  delete_indices();

  if (govfile)
    print_govs();

  delete_traces();

  if (PrintRules)
    add_rules();
  else if (govfile == NULL)
    print();
}
